# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import absolute_import, division, print_function

from flask import render_template

from datacube_ows.data import json_response
from datacube_ows.ogc_exceptions import WCS1Exception
from datacube_ows.ogc_utils import cache_control_headers, get_service_base_url
from datacube_ows.ows_configuration import get_config
from datacube_ows.query_profiler import QueryProfiler
from datacube_ows.utils import log_call
from datacube_ows.wcs1_utils import WCS1GetCoverageRequest, get_coverage_data

WCS_REQUESTS = ("DESCRIBECOVERAGE", "GETCOVERAGE")


@log_call
def handle_wcs1(nocase_args):
    operation = nocase_args.get("request", "").upper()
    if not operation:
        raise WCS1Exception("No operation specified", locator="Request parameter")
    elif operation == "GETCAPABILITIES":
        return get_capabilities(nocase_args)
    elif operation == "DESCRIBECOVERAGE":
        return desc_coverages(nocase_args)
    elif operation == "GETCOVERAGE":
        return get_coverage(nocase_args)
    else:
        raise WCS1Exception("Unrecognised operation: %s" % operation, locator="Request parameter")


@log_call
def get_capabilities(args):
    # TODO: Handle updatesequence request parameter for cache consistency.
    section = args.get("section")
    if section:
        section = section.lower()
    show_service = False
    show_capability = False
    show_content_metadata = False
    if section is None or section == "/":
        show_service = True
        show_capability = True
        show_content_metadata = True
    elif section == "/wcs_capabilities/service":
        show_service = True
    elif section == "/wcs_capabilities/capability":
        show_capability = True
    elif section == "/wcs_capabilities/contentmetadata":
        show_content_metadata = True
    else:
        raise WCS1Exception("Invalid section: %s" % section,
                            WCS1Exception.INVALID_PARAMETER_VALUE,
                            locator="Section parameter")

    # Extract layer metadata from Datacube.
    cfg = get_config()
    url = args.get('Host', args['url_root'])
    base_url = get_service_base_url(cfg.allowed_urls, url)
    headers = cache_control_headers(cfg.wms_cap_cache_age)
    headers["Content-Type"] = "application/xml"
    return (
        render_template("wcs_capabilities.xml",
                        show_service=show_service,
                        show_capability=show_capability,
                        show_content_metadata=show_content_metadata,
                        cfg=cfg,
                        base_url=base_url),
        200,
        cfg.response_headers(headers)
    )


@log_call
def desc_coverages(args):
    # Extract layer metadata from Datacube.
    cfg = get_config()

    coverages = args.get("coverage")
    products = []
    if coverages:
        coverages = coverages.split(",")
        for c in coverages:
            p = cfg.product_index.get(c)
            if p and p.wcs:
                products.append(p)
            else:
                raise WCS1Exception("Invalid coverage: %s" % c,
                                    WCS1Exception.COVERAGE_NOT_DEFINED,
                                    locator="Coverage parameter")
    else:
        for p in cfg.product_index.values():
            if p.ready and p.wcs:
                products.append(p)
    min_cache_age = min(p.resource_limits.wcs_desc_cache_rule for p in products)
    headers = cache_control_headers(min_cache_age)
    headers["Content-Type"] = "application/xml"
    return (
        render_template("wcs_desc_coverage.xml",
                        cfg=cfg,
                        products=products),
        200,
        cfg.response_headers(headers)
    )


@log_call
def get_coverage(args):
    cfg = get_config()
    req = WCS1GetCoverageRequest(args)
    qprof = QueryProfiler(req.ows_stats)
    n_datasets, data = get_coverage_data(req, qprof)
    if req.ows_stats:
        return json_response(qprof.profile())
    headers = {
        "Content-Type": req.format.mime,
        'content-disposition': 'attachment; filename=%s.%s' % (req.product_name, req.format.extension)
    }
    headers.update(req.product.resource_limits.wcs_cache_rules.cache_headers(n_datasets))
    return (
        req.format.renderer(req.version)(req, data),
        200,
        cfg.response_headers({
            "Content-Type": req.format.mime,
            'content-disposition': 'attachment; filename=%s.%s' % (req.product_name, req.format.extension)
        })
    )



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import io
import logging
from collections import defaultdict
from decimal import ROUND_HALF_UP, Decimal
from math import isclose
from typing import (Any, Hashable, List, MutableMapping, Optional, Tuple,
                    Union, cast)

import matplotlib
import numpy
from colour import Color
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, to_hex

try:
    from numpy.typing import NDArray
except ImportError:
    NDArray = numpy.ndarray

from xarray import Dataset

from datacube_ows.config_utils import CFG_DICT, OWSMetadataConfig
from datacube_ows.ogc_utils import ConfigException, FunctionWrapper
from datacube_ows.styles.base import StyleDefBase
from datacube_ows.styles.expression import Expression

_LOG = logging.getLogger(__name__)
matplotlib.use('Agg')

RAMP_SPEC = List[CFG_DICT]

UNSCALED_DEFAULT_RAMP = cast(RAMP_SPEC,
                             [
                                {
                                    "value": -1e-24,
                                    "color": "#000080",
                                    "alpha": 0.0
                                },
                                {
                                    "value": 0.0,
                                    "color": "#000080",
                                },
                                {
                                    "value": 0.1,
                                    "color": "#0000FF",
                                },
                                {
                                    "value": 0.3,
                                    "color": "#00FFFF",
                                },
                                {
                                    "value": 0.5,
                                    "color": "#00FF00",
                                },
                                {
                                    "value": 0.7,
                                    "color": "#FFFF00",
                                },
                                {
                                    "value": 0.9,
                                    "color": "#FF0000",
                                },
                                {
                                    "value": 1.0,
                                    "color": "#800000",
                                },
                             ]
)


def scale_unscaled_ramp(rmin: Union[int, float, str], rmax: Union[int, float, str], unscaled: RAMP_SPEC) -> RAMP_SPEC:
    """
    Take a unscaled (normalised) ramp that covers values from 0.0 to 1.0 and scale it linearly to cover the
    provided range.

    :param rmin: The new minimum value for the ramp range.
    :param rmax: The new maximum value for the ramp range.
    :param unscaled: The unscaled (normalised) ramp.
    :return: The scaled ramp.
    """
    if isinstance(rmin, float):
        nmin: float = rmin
    else:
        nmin = float(rmin)
    if isinstance(rmax, float):
        nmax: float = rmax
    else:
        nmax = float(rmax)
    return [
        {
            # pyre-ignore[6]
            "value": (nmax - nmin) * cast(float, u["value"]) + nmin,
            "color": u["color"],
            "alpha": u.get("alpha", 1.0)
        } for u in unscaled
    ]


def crack_ramp(ramp: RAMP_SPEC) -> Tuple[
    List[float],
    List[float], List[float],
    List[float], List[float],
]:
    """
    Split a colour ramp into separate (input) value and (output) RGBA lists.

    :param ramp: input (scaled) colour-ramp definition
    :return: A tuple of four lists of floats: representing values, red, green, blue, alpha.
    """
    values = cast(List[float], [])
    red = cast(List[float], [])
    green = cast(List[float], [])
    blue = cast(List[float], [])
    alpha = cast(List[float], [])
    for r in ramp:
        if isinstance(r["value"], float):
            value: float = cast(float, r["value"])
        else:
            value = float(cast(Union[int, str], r["value"]))
        values.append(value)
        color = Color(r["color"])
        red.append(color.red)
        green.append(color.green)
        blue.append(color.blue)
        alpha.append(float(cast(Union[float, int, str], r.get("alpha", 1.0))))

    return values, red, green, blue, alpha


def read_mpl_ramp(mpl_ramp: str) -> RAMP_SPEC:
    """
    Extract a named colour ramp from Matplotlib as a normalised OWS-compatible ramp specification

    :param mpl_ramp: The name of Matplotlib colour ramp
    :return: A normalised ramp specification.
    """
    unscaled_cmap = cast(RAMP_SPEC, [])
    try:
        cmap = plt.get_cmap(mpl_ramp)
    except:
        raise ConfigException(f"Invalid Matplotlib name: {mpl_ramp}")
    val_range = numpy.arange(0.1, 1.1, 0.1)
    rgba_hex = to_hex(cmap(0.0))
    unscaled_cmap.append(
        {
            "value": 0.0,
            "color": rgba_hex,
            "alpha": 1.0
        }
    )
    for val in val_range:
        rgba_hex = to_hex(cmap(val))
        unscaled_cmap.append(
            {
                "value": float(val),
                "color": rgba_hex
            }
        )
    return unscaled_cmap


class ColorRamp:
    """
    Represents a colour ramp for image and legend rendering purposes
    """
    def __init__(self, style: StyleDefBase,
                       ramp_cfg: CFG_DICT,
                       legend: "RampLegendBase") -> None:
        """
        :param style: The style owning the ramp
        :param ramp_cfg: Style config
        """
        self.style = style
        if "color_ramp" in ramp_cfg:
            raw_scaled_ramp = ramp_cfg["color_ramp"]
        else:
            rmin, rmax = cast(List[float], ramp_cfg["range"])
            unscaled_ramp = UNSCALED_DEFAULT_RAMP
            if "mpl_ramp" in ramp_cfg:
                unscaled_ramp = read_mpl_ramp(cast(str, ramp_cfg["mpl_ramp"]))
            raw_scaled_ramp = scale_unscaled_ramp(rmin, rmax, unscaled_ramp)
        self.ramp = raw_scaled_ramp

        self.values = cast(List[float], [])
        self.components = cast(MutableMapping[str, List[float]], {})
        self.crack_ramp()

        # Handle the mutual interdepencies between the ramp and the legend
        # 1. Let legend read its defaults from this ramp if needed
        legend.register_ramp(self)
        # 2. Extend our colour ramp to support legend if needed
        if self.style.auto_legend:
            fleg_begin = float(legend.begin)
            fleg_end = float(legend.end)
            leg_begin_in_ramp = False
            leg_end_in_ramp = False
            leg_begin_before_idx = None
            leg_end_before_idx = None
            for idx, col_point in enumerate(self.ramp):
                col_val = col_point["value"]
                if not leg_begin_in_ramp and leg_begin_before_idx is None:
                    if isclose(col_val, fleg_begin, abs_tol=1e-9):
                        leg_begin_in_ramp = True
                    elif col_val > fleg_begin:
                        leg_begin_before_idx = idx
                if not leg_end_in_ramp and leg_end_before_idx is None:
                    if isclose(col_val, fleg_end, abs_tol=1e-9):
                        end_in_ramp = True
                    elif col_val > fleg_end:
                        end_before_idx = idx
            if not leg_begin_in_ramp:
                color, alpha = self.color_alpha_at(fleg_begin)
                begin_col_point = {
                    "value": fleg_begin,
                    "color": color.get_hex(),
                    "alpha": alpha
                }
                if leg_begin_before_idx is None:
                    self.ramp.append(begin_col_point)
                else:
                    self.ramp.insert(leg_begin_before_idx, begin_col_point)
                if leg_end_before_idx is not None:
                    leg_end_before_idx += 1
            if not leg_end_in_ramp:
                color, alpha = self.color_alpha_at(fleg_end)
                end_col_point = {
                    "value": fleg_end,
                    "color": color.get_hex(),
                    "alpha": alpha
                }
                if leg_end_before_idx is None:
                    self.ramp.append(end_col_point)
                else:
                    self.ramp.insert(leg_end_before_idx, end_col_point)
            if not leg_end_in_ramp or not leg_begin_in_ramp:
                self.crack_ramp()

    def crack_ramp(self) -> None:
        values, r, g, b, a = crack_ramp(self.ramp)
        self.values = values
        self.components = {
            "red": r,
            "green": g,
            "blue": b,
            "alpha": a
        }

    def get_value(self, data: Union[float, "xarray.DataArray"], band: str) -> NDArray:
        return numpy.interp(data, self.values, self.components[band])

    def get_8bit_value(self, data: "xarray.DataArray", band: str) -> NDArray:
        val: NDArray = self.get_value(data, band)
        val = cast(NDArray, val * 255)
        return val.astype("uint8")

    def apply(self, data: "xarray.DataArray") -> "xarray.Dataset":
        imgdata = cast(MutableMapping[Hashable, Any], {})
        for band in self.components:
            imgdata[band] = (data.dims, self.get_8bit_value(data, band))
        imgdataset = Dataset(imgdata, coords=data.coords)
        return imgdataset

    def color_alpha_at(self, val: float) -> Tuple[Color, float]:
        color = Color(
            rgb=(
                self.get_value(val, "red").item(),
                self.get_value(val, "green").item(),
                self.get_value(val, "blue").item(),
            )
        )
        alpha = cast(float, self.get_value(val, "alpha"))
        return color, alpha


class RampLegendBase(StyleDefBase.Legend, OWSMetadataConfig):
    METADATA_ABSTRACT: bool = False
    METADATA_LEGEND_UNITS: bool = True
    METADATA_TICK_LABELS: bool = True

    def __init__(self, style_or_mdh: Union["StyleDefBase", "StyleDefBase.Legend"], cfg: CFG_DICT) -> None:
        super().__init__(style_or_mdh, cfg)
        raw_cfg = cast(CFG_DICT, self._raw_cfg)
        # Range - defaults deferred until we have parsed the associated ramp
        if "begin" not in raw_cfg:
            self.begin = Decimal("nan")
        else:
            self.begin = Decimal(cast(Union[str, float, int], raw_cfg["begin"]))
        if "end" not in raw_cfg:
            self.end = Decimal("nan")
        else:
            self.end = Decimal(cast(Union[str, float, int], raw_cfg["end"]))

        # decimal_places, rounder
        def rounder_str(prec: int) -> str:
            rstr = "1"
            if prec == 0:
                return rstr
            rstr += "."
            for i in range(prec - 1):
                rstr += "0"
            rstr += "1"
            return rstr

        self.decimal_places = raw_cfg.get("decimal_places", 1)
        if self.decimal_places < 0:
            raise ConfigException("decimal_places cannot be negative")
        self.rounder = Decimal(rounder_str(self.decimal_places))

        # Ticks - Non-explicit tick values deferred until we have parsed the associated ramp
        ticks_handled = False
        self.ticks_every: Optional[Decimal] = None
        self.tick_count: Optional[int] = None
        self.ticks: List[Decimal] = []
        if "ticks_every" in raw_cfg:
            if "tick_count" in raw_cfg:
                raise ConfigException("Cannot use tick count and ticks_every in the same legend")
            if "ticks" in raw_cfg:
                raise ConfigException("Cannot use ticks and ticks_every in the same legend")
            self.ticks_every = Decimal(cast(Union[int, float, str], raw_cfg["ticks_every"]))
            if self.ticks_every.is_zero() or self.ticks_every.is_signed():
                raise ConfigException("ticks_every must be greater than zero")
            ticks_handled = True
        if "ticks" in raw_cfg:
            if "tick_count" in raw_cfg:
                raise ConfigException("Cannot use tick count and ticks in the same legend")
            self.ticks = [Decimal(t) for t in cast(List[Union[str, int, float]], raw_cfg["ticks"])]
            ticks_handled = True
        if not ticks_handled:
            self.tick_count = int(cast(Union[str, int], raw_cfg.get("tick_count", 1)))
            if self.tick_count < 0:
                raise ConfigException("tick_count cannot be negative")
        # prepare for tick labels
        self.cfg_labels = cast(MutableMapping[str, MutableMapping[str, str]], raw_cfg.get("tick_labels", {}))
        defaults = self.cfg_labels.get("default", {})
        self.lbl_default_prefix = defaults.get("prefix", "")
        self.lbl_default_suffix = defaults.get("suffix", "")
        self.tick_labels: List[str] = []
        # handle matplotlib args
        self.strip_location = cast(List[float],
                                   raw_cfg.get("strip_location", [0.05, 0.5, 0.9, 0.15]))
        # throw error on legacy syntax
        self.fail_legacy()

    def fail_legacy(self) -> None:
        if any(
                legent in self._raw_cfg
                for legent in ["major_ticks", "offset", "scale_by", "radix_point"]
        ):
            raise ConfigException(
                f"Style {self.style.name} uses a no-longer supported format for legend configuration.  " +
                "Please refer to the documentation and update your config")

    def register_ramp(self, ramp: ColorRamp) -> None:
        if self.begin.is_nan():
            for col_def in ramp.ramp:
                if isclose(col_def.get("alpha", 1.0), 1.0, abs_tol=1e-9):
                    self.begin = Decimal(col_def["value"])
                    break
            if self.begin.is_nan():
                self.begin = Decimal(ramp.ramp[0]["value"])
        if self.end.is_nan():
            for col_def in reversed(ramp.ramp):
                if isclose(col_def.get("alpha", 1.0), 1.0, abs_tol=1e-9):
                    self.end = Decimal(col_def["value"])
                    break
            if self.end.is_nan():
                self.end = Decimal(ramp.ramp[-1]["value"])
        for t in self.ticks:
            if t < self.begin or t > self.end:
                raise ConfigException("Explicit ticks must all be within legend begin/end range")
        if self.ticks_every is not None:
            tickval = self.begin
            while tickval < self.end:
                self.ticks.append(tickval)
                tickval += self.ticks_every
            self.ticks.append(self.end)
        elif self.tick_count is not None:
            if self.tick_count == 0:
                self.ticks.append(self.begin)
            else:
                delta = self.end - self.begin
                dcount = Decimal(self.tick_count)
                for i in range(0, self.tick_count + 1):
                    tickval = self.begin + (Decimal(i) / dcount) * delta
                    self.ticks.append(tickval.quantize(self.rounder, rounding=ROUND_HALF_UP))

        # handle tick labels
        for tick in self.ticks:
            label_cfg = self.cfg_labels.get(str(tick))
            if label_cfg:
                prefix = label_cfg.get("prefix", self.lbl_default_prefix)
                suffix = label_cfg.get("suffix", self.lbl_default_suffix)
                label = label_cfg.get("label", str(tick))
                self.tick_labels.append(prefix + label + suffix)
            else:
                self.tick_labels.append(
                    self.lbl_default_prefix + str(tick) + self.lbl_default_suffix
                )
        self.parse_metadata(self._raw_cfg)

        # Check for legacy legend tips in ramp:
        for r in ramp.ramp:
            if "legend" in r:
                raise ConfigException(
                    f"Style {self.style.name} uses a no-longer supported format for legend configuration.  " +
                    "Please refer to the documentation and update your config")

    def tick_label(self, tick):
        try:
            tick_idx = self.ticks.index(tick)
            metaval = self.read_local_metadata(f"lbl_{tick}")
            if metaval:
                return metaval
            else:
                return self.tick_labels[tick_idx]
        except ValueError:
            _LOG.error("'%s' is a not a valid tick", tick)
            return None

    def create_cdict_ticks(self) -> Tuple[
        MutableMapping[str, List[Tuple[float, float, float]]],
        MutableMapping[float, str],
    ]:
        normalize_factor = float(self.end) - float(self.begin)
        cdict = cast(MutableMapping[str, List[Tuple[float, float, float]]], dict())
        bands = cast(MutableMapping[str, List[Tuple[float, float, float]]], defaultdict(list))
        started = False
        finished = False
        for index, ramp_point in enumerate(self.style_or_mdh.color_ramp.ramp):
            if finished:
                break

            value = cast(Union[float, int], ramp_point.get("value"))
            normalized = (value - float(self.begin)) / float(normalize_factor)

            if not started:
                if isclose(value, float(self.begin), abs_tol=1e-9):
                    started = True
                else:
                    continue
            if not finished:
                if isclose(value, float(self.end), abs_tol=1e-9):
                    finished = True

            for band, intensity in self.style_or_mdh.color_ramp.components.items():
                bands[band].append((normalized, intensity[index], intensity[index]))

        for band, blist in bands.items():
            cdict[band] = blist

        ticks = cast(MutableMapping[float, str], dict())
        for tick in self.ticks:
            value = float(tick)
            normalized = (value - float(self.begin)) / float(normalize_factor)
            ticks[normalized] = self.tick_label(tick)

        return cdict, ticks

    def display_title(self):
        if self.units:
            return f"{self.title}({self.units})"
        else:
            return self.title

    def plot_name(self):
        return f"{self.style.product.name}_{self.style.name}"

    def render(self, bytesio: io.BytesIO) -> None:
        cdict, ticks = self.create_cdict_ticks()
        plt.rcdefaults()
        if self.mpl_rcparams:
            plt.rcParams.update(self.mpl_rcparams)
        fig = plt.figure(figsize=(self.width, self.height))
        ax = fig.add_axes(self.strip_location)
        custom_map = LinearSegmentedColormap(self.plot_name(), cdict)
        color_bar = matplotlib.colorbar.ColorbarBase(
            ax,
            cmap=custom_map,
            orientation="horizontal")
        color_bar.set_ticks(list(ticks.keys()))
        color_bar.set_ticklabels(list(ticks.values()))
        color_bar.set_label(self.display_title())
        plt.savefig(bytesio, format='png')

    # For MetadataConfig
    @property
    def default_title(self) -> Optional[str]:
        return self.style.title



class ColorRampDef(StyleDefBase):
    """
    Colour ramp Style subclass
    """
    auto_legend = True

    def __init__(self,
                 product: "datacube_ows.ows_configuration.OWSNamedLayer",
                 style_cfg: CFG_DICT,
                 stand_alone: bool = False,
                 defer_multi_date: bool = False,
                 user_defined: bool = False) -> None:
        """"
        Constructor - refer to StyleDefBase
        """
        super(ColorRampDef, self).__init__(product, style_cfg,
                           stand_alone=stand_alone, defer_multi_date=True, user_defined=user_defined)
        style_cfg = cast(CFG_DICT, self._raw_cfg)
        self.color_ramp = ColorRamp(self, style_cfg, self.legend_cfg)
        self.include_in_feature_info = bool(style_cfg.get("include_in_feature_info", True))

        if "index_function" in style_cfg:
            self.index_function = FunctionWrapper(self,
                                                  cast(CFG_DICT, style_cfg["index_function"]),
                                                  stand_alone=self.stand_alone)
            if not self.stand_alone:
                for band in cast(List[str], style_cfg["needed_bands"]):
                    self.raw_needed_bands.add(band)
        elif "index_expression" in style_cfg:
            self.index_function = Expression(self, cast(str, style_cfg["index_expression"]))
            for band in self.index_function.needed_bands:
                self.raw_needed_bands.add(band)
            if self.stand_alone:
                self.needed_bands = [self.local_band(b) for b in self.raw_needed_bands]
                self.flag_bands = []
        else:
            raise ConfigException("Index function is required for index and hybrid styles. Style %s in layer %s" % (
                self.name,
                self.product.name
            ))
        if not defer_multi_date:
            self.parse_multi_date(style_cfg)

    def apply_index(self, data: "xarray.Dataset") -> "xarray.DataArray":
        """
        Caclulate index value across data.

        :param data: Input dataset
        :return: Matching dataarray carrying the index value
        """
        index_data = self.index_function(data)
        data['index_function'] = (index_data.dims, index_data.data)
        return data["index_function"]

    def transform_single_date_data(self, data: "xarray.Dataset") -> "xarray.Dataset":
        """
        Apply style to raw data to make an RGBA image xarray (single time slice only)

        :param data: Raw data, all bands.
        :return: RGBA uint8 xarray
        """
        d = self.apply_index(data)
        return self.color_ramp.apply(d)

    class Legend(RampLegendBase):
        pass

    class MultiDateHandler(StyleDefBase.MultiDateHandler):
        auto_legend = True

        def __init__(self, style: "ColorRampDef", cfg: CFG_DICT) -> None:
            """
            First stage initialisation

            :param style: The parent style object
            :param cfg: The multidate handler configuration
            """
            super().__init__(style, cfg)
            if self.animate:
                self.feature_info_label: Optional[str] = None
                self.color_ramp = style.color_ramp
            else:
                self.feature_info_label = cast(Optional[str], cfg.get("feature_info_label", None))
                self.color_ramp = ColorRamp(style, cfg, self.legend_cfg)

        def transform_data(self, data: "xarray.Dataset") -> "xarray.Dataset":
            """
            Apply image transformation

            :param data: Raw data
            :return: RGBA image xarray.  May have a time dimension
            """
            xformed_data = cast("ColorRampDef", self.style).apply_index(data)
            agg = self.aggregator(xformed_data)
            return self.color_ramp.apply(agg)

        class Legend(RampLegendBase):
            def plot_name(self):
                return f"{self.style.product.name}_{self.style.name}_{self.style_or_mdh.min_count}"

# Register ColorRampDef as Style subclass.
StyleDefBase.register_subclass(ColorRampDef, ("range", "color_ramp"))



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import absolute_import, division, print_function

import logging
import os
import warnings

import sentry_sdk
from botocore.credentials import RefreshableCredentials
from datacube.utils.aws import configure_s3_access
from flask import Flask, request
from flask_babel import Babel
from flask_log_request_id import RequestID, RequestIDLogFilter
from prometheus_flask_exporter.multiprocess import \
    GunicornInternalPrometheusMetrics
from rasterio.errors import NotGeoreferencedWarning
from sentry_sdk.integrations.flask import FlaskIntegration

from datacube_ows.ows_configuration import get_config

__all__ = [
    'initialise_babel',
    'initialise_logger',
    'initialise_ignorable_warnings',
    'initialise_debugging',
    'initialise_sentry',
    'initialise_aws_credentials',
    'parse_config_file',
    'initialise_flask',
    'initialise_prometheus',
    'generate_locale_selector',
    'CredentialManager',
]

def initialise_logger(name=None):
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('[%(asctime)s] [%(levelname)s] %(message)s'))
    handler.addFilter(RequestIDLogFilter())
    _LOG = logging.getLogger(name)
    _LOG.addHandler(handler)
    # If invoked using Gunicorn, link our root logger to the gunicorn logger
    # this will mean the root logs will be captured and managed by the gunicorn logger
    # allowing you to set the gunicorn log directories and levels for logs
    # produced by this application
    _LOG.setLevel(logging.getLogger('gunicorn.error').getEffectiveLevel())
    return _LOG


def initialise_ignorable_warnings():
    # Suppress annoying rasterio warning message every time we write to a non-georeferenced image format
    warnings.simplefilter("ignore", category=NotGeoreferencedWarning)


def initialise_debugging(log=None):
    # PYCHARM Debugging
    if os.environ.get("PYDEV_DEBUG"):
        import pydevd_pycharm
        pydevd_pycharm.settrace('172.17.0.1', port=12321, stdoutToServer=True, stderrToServer=True)
        if log:
            log.info("PyCharm Debugging enabled")


def initialise_sentry(log=None):
    if os.environ.get("SENTRY_KEY") and os.environ.get("SENTRY_PROJECT"):
        SENTRY_ENV_TAG = os.environ.get("SENTRY_ENV_TAG") if os.environ.get("SENTRY_ENV_TAG") else "dev"
        sentry_sdk.init(
            dsn="https://%s@sentry.io/%s" % (os.environ["SENTRY_KEY"], os.environ["SENTRY_PROJECT"]),
            environment=SENTRY_ENV_TAG,
            integrations = [FlaskIntegration()]
        )
        if log:
            log.info("Sentry initialised")

class CredentialManager:
    _instance = None

    def __new__(cls, log=None):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, log=None):
        # Startup initialisation of libraries controlled by environment variables
        self.use_aws = False
        self.unsigned = False
        self.requester_pays = False
        self.credentials = None
        self.log = log

        if log:
            log.debug("Initialising CredentialManager")

        # Boto3/AWS
        if os.environ.get("AWS_DEFAULT_REGION") or os.environ.get("AWS_REGION"):
            if "AWS_NO_SIGN_REQUEST" in os.environ:
                env_nosign = os.environ["AWS_NO_SIGN_REQUEST"]
                if env_nosign.lower() in ("y", "t", "yes", "true", "1"):
                    unsigned = True
                    # Workaround for rasterio bug
                    os.environ["AWS_NO_SIGN_REQUEST"] = "yes"
                    os.environ["AWS_ACCESS_KEY_ID"] = "fake"
                    os.environ["AWS_SECRET_ACCESS_KEY"] = "fake"
                else:
                    unsigned = False
                    # delete env variable
                    del os.environ["AWS_NO_SIGN_REQUEST"]
            else:
                unsigned = False
                if log:
                    log.warning("AWS_NO_SIGN_REQUEST is not set. " +
                                "The default behaviour has recently changed to False (i.e. signed requests) " +
                                "Please explicitly set $AWS_NO_SIGN_REQUEST to 'no' for unsigned requests.")
            env_requester_pays = os.environ.get("AWS_REQUEST_PAYER", "")
            requester_pays = False
            if env_requester_pays.lower() == "requester":
                requester_pays = True
            self.use_aws = True
            if log:
                if unsigned:
                    log.info("S3 access configured with unsigned requests")
                else:
                    log.info("S3 access configured with signed requests")
            self.unsigned = unsigned
            self.requester_pays = requester_pays
            self.renew_creds()

            if "AWS_S3_ENDPOINT" in os.environ and os.environ["AWS_S3_ENDPOINT"] == "":
                del os.environ["AWS_S3_ENDPOINT"]
        elif log:
            log.warning(
                "Environment variable $AWS_DEFAULT_REGION not set.  (This warning can be ignored if all data is stored locally.)")

    def _check_cred(self):
        if self.credentials and isinstance(self.credentials, RefreshableCredentials):
            if self.credentials.refresh_needed():
                self.renew_creds()
            elif self.log:
                # pylint: disable=protected-access
                self.log.info("Credentials look OK: %s seconds remaining", str(self.credentials._seconds_remaining()))
        elif self.log:
            self.log.debug("Credentials of type %s - NOT RENEWING", self.credentials.__class__.__name__)

    @classmethod
    def check_cred(cls):
        # pylint: disable=protected-access
        cls._instance._check_cred()

    def renew_creds(self):
        if self.use_aws:
            if self.log:
                self.log.info("Establishing/renewing credentials")
            self.credentials = configure_s3_access(aws_unsigned=self.unsigned,
                                                            requester_pays=self.requester_pays)
            if self.log:
                if isinstance(self.credentials, RefreshableCredentials):
                    # pylint: disable=protected-access
                    self.log.debug("%s seconds remaining", str(self.credentials._seconds_remaining()))


def initialise_aws_credentials(log=None):
    # pylint: disable=protected-access
    if CredentialManager._instance is None:
        cm = CredentialManager(log)


def parse_config_file(log=None):
    # Cache a parsed config file object
    # (unless deferring to first request)
    cfg = None
    if not os.environ.get("DEFER_CFG_PARSE"):
        cfg = get_config()
    return cfg


def initialise_flask(name):
    app = Flask(name.split('.')[0])
    RequestID(app)
    return app

def pass_through(undecorated):
    def decorator(*args, **kwargs):
        return undecorated(*args, **kwargs)
    decorator.__name__ = undecorated.__name__
    return decorator

class FakeMetrics:
    def do_not_track(self):
        return pass_through
    def counter(self, *args, **kwargs):
        return pass_through
    def histogram(self, *args, **kwargs):
        return pass_through
    def gauge(self, *args, **kwargs):
        return pass_through
    def summary(self, *args, **kwargs):
        return pass_through

def initialise_prometheus(app, log=None):
    # Prometheus
    if os.environ.get("prometheus_multiproc_dir", False):
        metrics = GunicornInternalPrometheusMetrics(app)
        if log:
            log.info("Prometheus metrics enabled")
        return metrics
    return FakeMetrics()

def request_extractor():
    qreq = request.args.get('request')
    return qreq

def generate_locale_selector(locales):
    def selector_template():
        return request.accept_languages.best_match(locales)
    return selector_template

def initialise_babel(cfg, app):
    if cfg and cfg.internationalised:
        app.config["BABEL_TRANSLATION_DIRECTORIES"] = cfg.translations_dir
        babel = Babel(app,
                      default_locale=cfg.locales[0],
                      default_domain=cfg.message_domain,
                      configure_jinja=False
                      )
        babel.localeselector(generate_locale_selector(cfg.locales))
        return babel
    else:
        return None



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import io
import logging
from datetime import datetime
from typing import Callable, List, MutableMapping, Optional, Union, cast

import numpy
import xarray
from colour import Color
from datacube.utils.masking import make_mask
from matplotlib import patches as mpatches
from matplotlib import pyplot as plt
from xarray import DataArray, Dataset

from datacube_ows.config_utils import (CFG_DICT, AbstractMaskRule,
                                       ConfigException, OWSMetadataConfig)
from datacube_ows.styles.base import StyleDefBase

_LOG = logging.getLogger(__name__)


class AbstractValueMapRule(AbstractMaskRule):
    """
    A Value Map Rule.

    Construct a ValueMap rule-set with ValueMapRule.value_map_from_config
    """
    def __init__(self, style_def: "ColorMapStyleDef", band: str, cfg: CFG_DICT) -> None:
        """
        Construct a Value Map Rule

        :param style_def: The owning ColorMapStyleDef object
        :param band: The name of the flag-band the rules apply to
        :param cfg: The rule specification
        """
        self.style = style_def
        super().__init__(band, cfg, mapper=style_def.local_band)
        cfg = cast(CFG_DICT, self._raw_cfg)

        self.title = cast(str, cfg["title"])
        self.abstract = cast(str, cfg.get("abstract"))
        if self.title and self.abstract:
            self.label: Optional[str] = f"{self.title} - {self.abstract}"
        elif self.title:
            self.label = self.title
        elif self.abstract:
            self.label = self.abstract
        else:
            self.label = None
        self.parse_color(cfg)

    @property
    def context(self) -> str:
        return f"style {self.style.name} in layer {self.style.product.name} valuemap rule"

    def parse_color(self, cfg: CFG_DICT):
        self.color_str = cast(str, cfg["color"])
        self.rgb = Color(self.color_str)
        if cfg.get("mask"):
            self.alpha = 0.0
        else:
            self.alpha = float(cast(Union[float, int, str], cfg.get("alpha", 1.0)))

    @classmethod
    def value_map_from_config(cls,
                          style_or_mdh: Union["ColorMapStyleDef", "ColorMapStyleDef.MultiDateHandler"],
                          cfg: CFG_DICT
                              ) -> MutableMapping[str, List["AbstractValueMapRule"]]:
        """
        Create a multi-date value map rule set from a config specification

        :param style: The parent style definition object
        :param cfg: The specification for the multi-date value map.

        :return: A value map ruleset dictionary.
        """
        if isinstance(style_or_mdh, ColorMapStyleDef):
            typ = ValueMapRule
        else:
            mdh = cast(ColorMapStyleDef.MultiDateHandler, style_or_mdh)
            if mdh.aggregator:
                style_or_mdh = mdh.style
                typ = ValueMapRule
            else:
                if mdh.min_count != mdh.max_count:
                    raise ConfigException(
                        "MultiDate value map only supported on multi-date handlers with min_count and max_count equal.")
                typ = MultiDateValueMapRule
        vmap: MutableMapping[str, List["AbstractValueMapRule"]] = {}
        for band_name, rules in cfg.items():
            band_rules = [typ(style_or_mdh, band_name, rule) for rule in cast(List[CFG_DICT], rules)]
            vmap[band_name] = band_rules
        return vmap


class ValueMapRule(AbstractValueMapRule):
    """
    A Value Map Rule.

    Construct a ValueMap rule-set with ValueMapRule.value_map_from_config
    """
    def __init__(self, style_cfg: "ColorMapStyleDef", band: str,
                 cfg: CFG_DICT) -> None:
        """
        Construct a Multi-date Value Map Rule

        :param mdh: The owning ColorMapStyleDef object
        :param band: The name of the flag-band the rules apply to
        :param cfg: The rule specification
        """
        super().__init__(style_def=style_cfg, band=band, cfg=cfg)


class MultiDateValueMapRule(AbstractValueMapRule):
    """
    A  Multi-Date Value Map Rule.

    Construct a Multi-Date ValueMap rule-set with MultiDateValueMapRule.value_map_from_config
    """
    def __init__(self, mdh: "ColorMapStyleDef.MultiDateHandler", band: str,
                 cfg: CFG_DICT) -> None:
        """
        Construct a Multi-date Value Map Rule

        :param mdh: The owning ColorMapStyleDef object
        :param band: The name of the flag-band the rules apply to
        :param cfg: The rule specification
        """
        self.mdh = mdh
        self.invert: List[bool] = []
        self.flags: Optional[List[CFG_DICT]] = []
        self.or_flags: Optional[List[bool]] = []
        self.values: Optional[List[List[int]]] = []
        super().__init__(style_def=mdh.style, band=band, cfg=cfg)

    def parse_rule_spec(self, cfg: CFG_DICT):
        if "invert" in cfg:
            self.invert = [bool(b) for b in cfg["invert"]]
        else:
            self.invert = [False] * self.mdh.max_count
        if len(self.invert) != self.mdh.max_count:
            raise ConfigException(f"Invert entry has wrong number of rule sets for date count")
        if "flags" in cfg:
            date_flags = cast(CFG_DICT, cfg["flags"])
            if len(date_flags) != self.mdh.max_count:
                raise ConfigException(f"Flags entry has wrong number of rule sets for date count")
            for flags in date_flags:
                or_flag: bool = False
                if "or" in flags and "and" in flags:
                    raise ConfigException(f"MultiDateValueMap rule in {self.mdh.style.name} of layer {self.mdh.style.product.name} combines 'and' and 'or' rules")
                elif "or" in flags:
                    or_flag = True
                    flags = cast(CFG_DICT, flags["or"])
                elif "and" in flags:
                    flags = cast(CFG_DICT, flags["and"])
                self.flags.append(flags)
                self.or_flags.append(or_flag)
        else:
            self.flags = None
            self.or_flags = None
        if "values" in cfg:
            self.values = cast(List[List[int]], list(cfg["values"]))
        else:
            self.values = None
        if not self.flags and not self.values:
            raise ConfigException(f"Multi-Date Value map rule in {self.context} must have a non-empty 'flags' or 'values' section.")
        if self.flags and self.values:
            raise ConfigException(f"Multi-Date Value map rule in {self.context} has both a 'flags' and a 'values' section - choose one.")

    def create_mask(self, data: DataArray) -> DataArray:
        """
        Create a mask from raw flag band data.

        :param data: Multi-date Raw flag data, assumed to be for this rule's flag band.
        :return: A boolean dateless DataArray, True where the data matches this rule
        """
        date_slices = (data.sel(time=dt) for dt in data.coords["time"].values)
        mask: Optional[DataArray] = None
        if self.values:
            for d_slice, vals, invert in zip(date_slices, self.values, self.invert):
                d_mask: Optional[DataArray] = None
                if len(vals) == 0:
                    d_mask = d_slice == d_slice
                else:
                    for v in cast(List[int], vals):
                        vmask = d_slice == v
                        if d_mask is None:
                            d_mask = vmask
                        else:
                            d_mask |= vmask
                if d_mask is not None and invert:
                    d_mask = ~d_mask # pylint: disable=invalid-unary-operand-type
                if mask is None:
                    mask = d_mask
                else:
                    mask &= d_mask
        else:
            for d_slice, flags, or_flags, invert in zip(date_slices, self.flags, self.or_flags, self.invert):
                d_mask: Optional[DataArray] = None
                if not flags:
                    d_mask = d_slice == d_slice
                elif or_flags:
                    for f in cast(CFG_DICT, flags).items():
                        f = {f[0]: f[1]}
                        if d_mask is None:
                            d_mask = make_mask(d_slice, **f)
                        else:
                            d_mask |= make_mask(d_slice, **f)
                else:
                    d_mask = make_mask(d_slice, **cast(CFG_DICT, flags))
                if invert:
                    d_mask = ~d_mask # pylint: disable=invalid-unary-operand-type
                if mask is None:
                    mask = d_mask
                else:
                    mask &= d_mask
        return mask


def convert_to_uint8(fval):
    scaled = int(fval * 255.0 + 0.5)
    clipped = min(max(scaled, 0), 255)
    return clipped


def apply_value_map(value_map: MutableMapping[str, List[AbstractValueMapRule]],
                    data: Dataset,
                    band_mapper: Callable[[str], str]) -> Dataset:
    imgdata = Dataset(coords={k: v for k, v in data.coords.items() if k != "time"})
    shape = list(imgdata.dims.values())
    for channel in ("red", "green", "blue", "alpha"):
        c = numpy.full(shape, 0, dtype="uint8")
        imgdata[channel] = DataArray(c, coords=imgdata.coords)
    for cfg_band, rules in value_map.items():
        # Run through each item
        band = band_mapper(cfg_band)
        bdata = cast(DataArray, data[band])
        if bdata.dtype.kind == 'f':
            # Convert back to int for bitmasking
            bdata = ColorMapStyleDef.reint(bdata)
        for rule in reversed(rules):
            mask = rule.create_mask(bdata)
            if mask.data.any():
                for channel in ("red", "green", "blue", "alpha"):
                    if channel == "alpha":
                        val = convert_to_uint8(rule.alpha)
                    else:
                        val = convert_to_uint8(getattr(rule.rgb, channel))
                    imgdata[channel] = xarray.where(mask, val, imgdata[channel])
    return imgdata


class PatchTemplate:
    def __init__(self, idx: int, rule: AbstractValueMapRule) -> None:
        self.idx = idx
        self.colour = rule.rgb.hex_l
        self.label = rule.label


class ColorMapLegendBase(StyleDefBase.Legend, OWSMetadataConfig):
    METADATA_ABSTRACT: bool = False
    METADATA_VALUE_RULES: bool = True

    def __init__(self, style_or_mdh: Union["StyleDefBase", "StyleDefBase.Legend"], cfg: CFG_DICT) -> None:
        super().__init__(style_or_mdh, cfg)
        raw_cfg = cast(CFG_DICT, self._raw_cfg)
        self.ncols = int(raw_cfg.get("ncols", 1))
        if self.ncols < 1:
            raise ConfigException("ncols must be a positive integer")
        self.patches: List[PatchTemplate] = []

    def register_value_map(self, value_map: MutableMapping[str, List["AbstractValueMapRule"]]) -> None:
        for band in value_map.keys():
            for idx, rule in reversed(list(enumerate(value_map[band]))):
                # only include values that are not transparent (and that have a non-blank title or abstract)
                if rule.alpha > 0.001 and rule.label:
                    self.patches.append(PatchTemplate(idx, rule))
        self.parse_metadata(self._raw_cfg)

    def render(self, bytesio: io.BytesIO) -> None:
        patches = [
            mpatches.Patch(color=pt.colour, label=self.patch_label(pt.idx))
            for pt in self.patches
        ]
        plt.rcdefaults()
        if self.mpl_rcparams:
            plt.rcParams.update(self.mpl_rcparams)
        plt.figure(figsize=(self.width, self.height))
        plt.axis('off')
        if self.title:
            plt.legend(handles=patches,
                       loc='center',
                       ncol=self.ncols,
                       frameon=False,
                       title=self.title)
        else:
            plt.legend(handles=patches,
                       loc='center',
                       ncol=self.ncols,
                       frameon=False)
        plt.savefig(bytesio, format='png')

    def patch_label(self, idx: int):
        return self.read_local_metadata(f"rule_{idx}")

    # For MetadataConfig
    @property
    def default_title(self) -> Optional[str]:
        return ""



class ColorMapStyleDef(StyleDefBase):
    """
    Style subclass for value-map styles
    """
    auto_legend = True

    def __init__(self,
                 product: "datacube_ows.ows_configuration.OWSNamedLayer",
                 style_cfg: CFG_DICT,
                 stand_alone: bool = False,
                 user_defined: bool = False) -> None:
        """"
        Constructor - refer to StyleDefBase
        """
        super().__init__(product, style_cfg, stand_alone=stand_alone, user_defined=user_defined)
        style_cfg = cast(CFG_DICT, self._raw_cfg)
        self.value_map = AbstractValueMapRule.value_map_from_config(self, cast(CFG_DICT, style_cfg["value_map"]))
        self.legend_cfg.register_value_map(self.value_map)
        for mdh in self.multi_date_handlers:
            mdh.legend_cfg.register_value_map(mdh.value_map)
        for band in self.value_map.keys():
            self.raw_needed_bands.add(band)

    @staticmethod
    def reint(data: DataArray) -> DataArray:
        """
        Convert a data-array to int.

        :param data: input data (potentially non-integer)
        :return: same data cast to integer
        """
        inted = data.astype("int")
        if hasattr(data, "attrs"):
            attrs = data.attrs
            inted.attrs = attrs
        return inted

    @staticmethod
    def create_colordata(data: DataArray, rgb: Color, alpha: float, mask: DataArray) -> Dataset:
        """Colour a mask with a given colour/alpha"""
        target = Dataset(coords=data.coords)
        colors = ["red", "green", "blue", "alpha"]
        for color in colors:
            val = alpha if color == "alpha" else getattr(rgb, color)
            c = numpy.full(data.shape, val)
            target[color] = DataArray(c, dims=data.dims, coords=data.coords)
        # pyre-ignore[6]
        masked = target.where(mask).where(numpy.isfinite(data))  # remask
        return masked

    def transform_single_date_data(self, data: Dataset) -> Dataset:
        """
        Apply style to raw data to make an RGBA image xarray (single time slice only)

        :param data: Raw data, all bands.
        :return: RGBA uint8 xarray
        """
        # pylint: disable=too-many-locals, too-many-branches
        # extent mask data per band to preseve nodata
        _LOG.debug("transform begin %s", datetime.now())
        # if extent_mask is not None:
        #    for band in data.data_vars:
        # try:
        #            data[band] = data[band].where(extent_mask, other=data[band].attrs['nodata'])
        #        except AttributeError:
        #            data[band] = data[band].where(extent_mask)
        return apply_value_map(self.value_map, data, self.product.band_idx.band)

    class Legend(ColorMapLegendBase):
        pass

    class MultiDateHandler(StyleDefBase.MultiDateHandler):
        auto_legend = True
        non_animate_requires_aggregator = False

        def __init__(self, style: "ColorMapStyleDef", cfg: CFG_DICT) -> None:
            """
            First stage initialisation

            :param style: The parent style object
            :param cfg: The multidate handler configuration
            """
            super().__init__(style, cfg)
            self._value_map: Optional[MutableMapping[str, AbstractValueMapRule]] = None
            if self.animate:
                if "value_map" in self._raw_cfg:
                    raise ConfigException("Multidate value maps not supported for animation handlers")
            else:
                self._value_map = AbstractValueMapRule.value_map_from_config(self,
                                                        cast(CFG_DICT, self._raw_cfg["value_map"]))

        @property
        def value_map(self):
            if self._value_map is None:
                self._value_map = self.style.value_map
            return self._value_map

        def transform_data(self, data: "xarray.Dataset") -> "xarray.Dataset":
            """
            Apply image transformation

            :param data: Raw data
            :return: RGBA image xarray.  May have a time dimension
            """
            if self.aggregator is None:
                return apply_value_map(self.value_map, data, self.style.product.band_idx.band)
            else:
                agg = self.aggregator(data)
                return apply_value_map(self.value_map, agg, self.style.product.band_idx.band)

        class Legend(ColorMapLegendBase):
            pass

# Register ColorMapStyleDef as a style subclass.
StyleDefBase.register_subclass(ColorMapStyleDef, "value_map")



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

from datacube_ows.styles.api.base import (  # noqa: F401 isort:skip
    StandaloneStyle, apply_ows_style, apply_ows_style_cfg,
    generate_ows_legend_style, generate_ows_legend_style_cfg,
    plot_image, plot_image_with_style, plot_image_with_style_cfg)

from datacube_ows.ogc_utils import create_geobox, xarray_image_as_png       # noqa: F401 isort:skip
from datacube_ows.band_utils import scale_data, scalable, band_modulator    # noqa: F401 isort:skip



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from functools import wraps
from time import monotonic
from typing import Any, Callable, TypeVar

F = TypeVar('F', bound=Callable[..., Any])

def log_call(func: F) -> F:
    """
    Profiling function decorator

    Placing @log_call at the top of a function or method, results in all calls to that function or method
    being logged at debug level.
    """
    @wraps(func)
    def log_wrapper(*args, **kwargs):
        _LOG = logging.getLogger()
        _LOG.debug("%s args: %s kwargs: %s", func.__name__, args, kwargs)
        return func(*args, **kwargs)
    return log_wrapper


def time_call(func: F) -> F:
    """
    Profiling function decorator

    Placing @log_call at the top of a function or method, results in all calls to that function or method
    being timed at debug level.

    For debugging or optimisation research only.  Should not occur in mainline code.
    """
    @wraps(func)
    def timing_wrapper(*args, **kwargs):
        start: float = monotonic()
        result: Any = func(*args, **kwargs)
        stop: float = monotonic()
        _LOG = logging.getLogger()
        _LOG.debug("%s took: %d ms", func.__name__, int((stop - start) * 1000))
        return result
    return timing_wrapper


def group_by_statistical() -> "datacube.api.query.GroupBy":
    """
    Returns an ODC GroupBy object, suitable for daily statistical/summary data.
    """
    from datacube.api.query import GroupBy

    return GroupBy(
        dimension='time',
        group_by_func=lambda ds: ds.time.begin,
        units='seconds since 1970-01-01 00:00:00',
        sort_key=lambda ds: ds.time.begin
    )


def get_sqlconn(dc: "datacube.Datacube") -> "sqlalchemy.engine.base.Connection":
    """
    Extracts a SQLAlchemy database connection from a Datacube object.

    :param dc: An initialised Datacube object
    :return: A SQLAlchemy database connection object.
    """
    # pylint: disable=protected-access
    return dc.index._db._engine.connect()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
from typing import (Any, Callable, Hashable, List, MutableMapping, Optional,
                    Union, cast)

import numpy as np
from xarray import DataArray, Dataset

from datacube_ows.config_utils import CFG_DICT
from datacube_ows.ogc_utils import ConfigException, FunctionWrapper
from datacube_ows.styles.base import StyleDefBase

# pylint: disable=abstract-method

LINEAR_COMP_DICT = MutableMapping[str, Union[float, List[float]]]


class ComponentStyleDef(StyleDefBase):
    """
    Style Subclass that allows the behaviour of each component (red, green, blue, alpha) to be
    specified independently.
    """
    def __init__(self, product: "datacube_ows.ows_configuration.OWSNamedLayer",
                 style_cfg: CFG_DICT,
                 stand_alone: bool = False,
                 defer_multi_date: bool = False,
                 user_defined: bool = False) -> None:
        """
        See superclass
        """
        super().__init__(product, style_cfg,
                         stand_alone=stand_alone, defer_multi_date=defer_multi_date, user_defined=user_defined)
        style_cfg: CFG_DICT = cast(CFG_DICT, self._raw_cfg)
        self.raw_rgb_components: MutableMapping[str, Union[Callable, LINEAR_COMP_DICT]] = {}
        raw_components = cast(MutableMapping[str, Union[Callable, CFG_DICT]], style_cfg["components"])
        for imgband in ["red", "green", "blue", "alpha"]:
            components = raw_components.get(imgband)
            if components is None:
                if imgband == "alpha":
                    continue
                else:
                    raise ConfigException(f"No components defined for {imgband} band in style {self.name}, layer {product.name}")
            elif callable(components) or "function" in components:
                self.raw_rgb_components[imgband] = FunctionWrapper(self.product, components,
                                                                   stand_alone=self.stand_alone)
                if not self.stand_alone:
                    if "additional_bands" not in style_cfg:
                        raise ConfigException(f"Style with a function component must declare additional_bands.")
                    for b in cast(List[str], style_cfg.get("additional_bands", [])):
                        self.raw_needed_bands.add(b)
            else:
                components = cast(LINEAR_COMP_DICT, components)
                self.raw_rgb_components[imgband] = components
                for k in components.keys():
                    if k != "scale_range":
                        self.raw_needed_bands.add(k)
        self.rgb_components = cast(MutableMapping[str, Union[None, Callable, LINEAR_COMP_DICT]], {})

        self.scale_factor = style_cfg.get("scale_factor")
        if "scale_range" in style_cfg:
            self.scale_min, self.scale_max = cast(List[Optional[float]], style_cfg["scale_range"])
        elif self.scale_factor:
            self.scale_min = 0.0
            self.scale_max = 255.0 * self.scale_factor
        else:
            self.scale_min = None
            self.scale_max = None

        self.component_scale_ranges: MutableMapping[str, MutableMapping[str, float]] = {}
        for cn, cd in raw_components.items():
            if not callable(cd) and "scale_range" in cd:
                scale_range = cast(List[float], cd["scale_range"])
                self.component_scale_ranges[cn] = {
                    "min": scale_range[0],
                    "max": scale_range[1],
                }
            else:
                self.component_scale_ranges[cn] = {
                    "min": self.scale_min,
                    "max": self.scale_max,
                }

    # pylint: disable=attribute-defined-outside-init
    def make_ready(self, dc, *args, **kwargs) -> None:
        """
        Second-phase (db aware) initialisation

        Mostly sorting out bands, esp flag bands.

        :param dc: A datacube object
        """
        self.rgb_components = cast(MutableMapping[str, Union[None, Callable, LINEAR_COMP_DICT]], {})
        for band, component in self.raw_rgb_components.items():
            if not component or callable(component):
                self.rgb_components[band] = component
            else:
                self.rgb_components[band] = self.dealias_components(component)
        super().make_ready(dc, *args, **kwargs)


        self.raw_rgb_components: MutableMapping[str, Union[Callable, LINEAR_COMP_DICT]] = {}

    def dealias_components(self, comp_in: Optional[LINEAR_COMP_DICT]) -> Optional[LINEAR_COMP_DICT]:
        """
        Convert a component dictionary with band aliases to a component dictionary using canonical band names.

        :param comp_in: A linear component dictionary with band aliases
        :return: Equivalent linear component dictionary with canonical band names.
        """
        if self.stand_alone:
            return comp_in
        elif comp_in is None:
            return None
        else:
            return {
                self.product.band_idx.band(self.local_band(band_alias)): value
                for band_alias, value in comp_in.items() if band_alias not in ['scale_range']
            }

    def compress_band(self, component_name: str, imgband_data: "xarray.DataArray") -> "xarray.DataArray":
        """
        Compress dynamic range of a component data array to uint8 range (0-255)

        :param component_name: The name of the component being compressed (i.e. 'red', 'green', 'blue' or 'alpha')
        :param imgband_data: The input, uncompressed DataArray for the component
        :return:
        """
        sc_min: float = self.component_scale_ranges[component_name]["min"]
        sc_max: float = self.component_scale_ranges[component_name]["max"]
        clipped = imgband_data.clip(sc_min, sc_max)
        normalized = (clipped - sc_min) / (sc_max - sc_min)
        return normalized * 255


    def transform_single_date_data(self, data: "xarray.Dataset") -> "xarray.Dataset":
        """
        Apply style to raw data to make an RGBA image xarray (single time slice only)

        :param data: Raw data, all bands.
        :return: RGBA uint8 xarray
        """
        imgdata = cast(MutableMapping[Hashable, Any], {})
        for imgband, components in self.rgb_components.items():
            if callable(components):
                imgband_data = components(data)
                imgband_data = imgband_data.astype('uint8')
                imgdata[imgband] = imgband_data
            else:
                imgband_data = None
                for band, intensity in components.items():
                    if callable(intensity):
                        imgband_component = intensity(data[band], band, imgband)
                    elif band == "scale_range":
                        continue
                    else:
                        imgband_component = data[band] * intensity

                    if imgband_data is not None:
                        imgband_data += imgband_component
                    else:
                        imgband_data = imgband_component
                if imgband_data is None:
                    null_np = np.zeros(list(data.dims.values()), 'uint8')
                    imgband_data = DataArray(null_np, data.coords, data.dims.keys())
                if imgband != "alpha":
                    imgband_data = self.compress_band(imgband, imgband_data)
                imgdata[imgband] = imgband_data.astype("uint8")

        image_dataset = Dataset(imgdata)
        return image_dataset


# Register ComponentStyleDef as a Style subclass.
StyleDefBase.register_subclass(ComponentStyleDef, "components")



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

#pylint: skip-file
import os
import sys

# This is the directory of the source code that the web app will run from
sys.path.append("/opt")

# The location of the datcube config file.
os.environ.setdefault("DATACUBE_CONFIG_PATH", "/opt/odc/.datacube.conf.local")

from datacube_ows.ogc import app  # isort:skip

application = app


def main():
    app.run()


if __name__ == '__main__':
    main()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import json
from enum import Enum
from typing import Any, Iterable, Optional, Tuple, Union, cast

from datacube.utils.geometry import Geometry as ODCGeom
from geoalchemy2 import Geometry
from psycopg2.extras import DateTimeTZRange
from sqlalchemy import SMALLINT, Column, MetaData, Table, or_, select, text
from sqlalchemy.dialects.postgresql import TSTZRANGE, UUID
from sqlalchemy.sql.functions import count


def get_sqlalc_engine(index: "datacube.index.Index") -> "sqlalchemy.engine.base.Engine":
    # pylint: disable=protected-access
    return index._db._engine


def get_st_view(meta: MetaData) -> Table:
    return Table('space_time_view', meta,
             Column('id', UUID()),
             Column('dataset_type_ref', SMALLINT()),
             Column('spatial_extent', Geometry(from_text='ST_GeomFromGeoJSON', name='geometry')),
             Column('temporal_extent', TSTZRANGE())
                 )
_meta = MetaData()
st_view = get_st_view(_meta)


class MVSelectOpts(Enum):
    """
    Enum for mv_search_datasets sel parameter.

    ALL: return all columns, select *, as result set
    IDS: return list of database_ids only.
    DATASETS: return list of ODC dataset objects
    COUNT: return a count of matching datasets
    EXTENT: return full extent of query result as a Geometry
    """
    ALL = 0
    IDS = 1
    COUNT = 2
    EXTENT = 3
    DATASETS = 4
    INVALID = 9999

    def sel(self, stv: Table) -> Iterable["sqlalchemy.sql.elements.ClauseElement"]:
        if self == self.ALL:
            return [stv]
        if self == self.IDS or self == self.DATASETS:
            return [stv.c.id]
        if self == self.COUNT:
            return [cast("sqlalchemy.sql.elements.ClauseElement", count(stv.c.id))]
        if self == self.EXTENT:
            return [text("ST_AsGeoJSON(ST_Union(spatial_extent))")]
        assert False


def mv_search(index: "datacube.index.Index",
              sel: MVSelectOpts = MVSelectOpts.IDS,
              times: Optional[Iterable[Tuple[datetime.datetime, datetime.datetime]]] = None,
              geom: Optional[ODCGeom] = None,
              products: Optional[Iterable["datacube.model.DatasetType"]] = None) -> Union[
        Iterable[Iterable[Any]],
        Iterable[str],
        Iterable["datacube.model.Dataset"],
        int,
        None,
        ODCGeom]:
    """
    Perform a dataset query via the space_time_view

    :param products: An iterable of combinable products to search
    :param index: A datacube index (required)

    :param sel: Selection mode - a MVSelectOpts enum. Defaults to IDS.
    :param times: A list of pairs of datetimes (with time zone)
    :param geom: A datacube.utils.geometry.Geometry object

    :return: See MVSelectOpts doc
    """
    engine = get_sqlalc_engine(index)
    stv = st_view
    if products is None:
        raise Exception("Must filter by product/layer")
    prod_ids = [p.id for p in products]

    s = select(sel.sel(stv)).where(stv.c.dataset_type_ref.in_(prod_ids))
    if times is not None:
        s = s.where(
            or_(
                *[
                    stv.c.temporal_extent.op("&&")(DateTimeTZRange(*t))
                    for t in times
                ]
            )
        )
    orig_crs = None
    if geom is not None:
        orig_crs = geom.crs
        if str(geom.crs) != "EPSG:4326":
            geom = geom.to_crs("EPSG:4326")
        geom_js = json.dumps(geom.json)
        s = s.where(stv.c.spatial_extent.intersects(geom_js))
    # print(s) # Print SQL Statement
    conn = engine.connect()
    if sel == MVSelectOpts.ALL:
        return conn.execute(s)
    if sel == MVSelectOpts.IDS:
        return [r[0] for r in conn.execute(s)]
    if sel in (MVSelectOpts.COUNT, MVSelectOpts.EXTENT):
        for r in conn.execute(s):
            if sel == MVSelectOpts.COUNT:
                return r[0]
            if sel == MVSelectOpts.EXTENT:
                geojson = r[0]
                if geojson is None:
                    return None
                uniongeom = ODCGeom(json.loads(geojson), crs="EPSG:4326")
                if geom:
                    intersect = uniongeom.intersection(geom)
                    if intersect.wkt == 'POLYGON EMPTY':
                        return None
                    if orig_crs and orig_crs != "EPSG:4326":
                        intersect = intersect.to_crs(orig_crs)
                else:
                    intersect = uniongeom
                return intersect
    if sel == MVSelectOpts.DATASETS:
        ids = [r[0] for r in conn.execute(s)]
        return index.datasets.bulk_get(ids)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import absolute_import, division, print_function

from flask import render_template

from datacube_ows.data import feature_info, get_map
from datacube_ows.legend_generator import legend_graphic
from datacube_ows.ogc_exceptions import WMSException
from datacube_ows.ogc_utils import cache_control_headers, get_service_base_url
from datacube_ows.ows_configuration import get_config
from datacube_ows.utils import log_call

WMS_REQUESTS = ("GETMAP", "GETFEATUREINFO", "GETLEGENDGRAPHIC")


@log_call
def handle_wms(nocase_args):
    operation = nocase_args.get("request", "").upper()
    # WMS operation Map
    if not operation:
        raise WMSException("No operation specified", locator="Request parameter")
    elif operation == "GETCAPABILITIES":
        return get_capabilities(nocase_args)
    elif operation == "GETMAP":
        return get_map(nocase_args)
    elif operation == "GETFEATUREINFO":
        return feature_info(nocase_args)
    elif operation == "GETLEGENDGRAPHIC":
        return legend_graphic(nocase_args)
    else:
        raise WMSException("Unrecognised operation: %s" % operation, WMSException.OPERATION_NOT_SUPPORTED,
                           "Request parameter")


@log_call
def get_capabilities(args):
    # TODO: Handle updatesequence request parameter for cache consistency.
    # Note: Only WMS v1.3.0 is fully supported at this stage, so no version negotiation is necessary
    # Extract layer metadata from Datacube.
    cfg = get_config()
    url = args.get('Host', args['url_root'])
    base_url = get_service_base_url(cfg.allowed_urls, url)
    headers = cache_control_headers(cfg.wms_cap_cache_age)
    headers["Content-Type"] = "application/xml"
    return (
        render_template(
            "wms_capabilities.xml",
            cfg=cfg,
            base_url=base_url),
        200,
        cfg.response_headers(headers)
    )



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2021 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import absolute_import

import io
import logging

import matplotlib
import numpy as np
# from flask import make_response
from PIL import Image

from datacube_ows.ogc_exceptions import WMSException
from datacube_ows.ogc_utils import resp_headers
from datacube_ows.wms_utils import GetLegendGraphicParameters

# Do not use X Server backend

matplotlib.use('Agg')

_LOG = logging.getLogger(__name__)


def legend_graphic(args):
    params = GetLegendGraphicParameters(args)
    img = create_legends_from_styles(params.styles,
                        ndates=len(params.times))
    if img is None:
        raise WMSException("No legend is available for this request", http_response=404)
    return img


def create_legend_for_style(product, style_name, ndates=0):
    if style_name not in product.style_index:
        return None
    style = product.style_index[style_name]
    return create_legends_from_styles([style], ndates)


def create_legends_from_styles(styles, ndates=0):
    # Run through all values in style cfg and generate
    imgs = []
    for s in styles:
        img = s.render_legend(ndates)
        if img is not None:
            imgs.append(img)

    if not imgs:
        return None
    min_shape = sorted([(np.sum(i.size), i.size) for i in imgs])[0][1]
    imgs_comb = np.vstack([np.asarray(i.resize(min_shape)) for i in imgs])
    imgs_comb = Image.fromarray(imgs_comb)
    b = io.BytesIO()
    imgs_comb.save(b, 'png')
    # legend = make_response(b.getvalue())
    # legend.mimetype = 'image/png'
    # b.close()
    return (b.getvalue(), 200, resp_headers({"Content-Type": "image/png"}))
